//----------------------------------------------------------------------------------------------------------------------
//
//  gRabbit - Graph Builder for Basic Implementation Tests
//  Copyright 2004, Nelson Guedes Paulo Junior (npaulo@linux.ime.usp.br)
//
//  Under gRabbit license (see LICENSE.TXT for details)
//
//
// - Animation Files Parser
//
//----------------------------------------------------------------------------------------------------------------------
#ifdef __GNUG__
  #pragma implementation
#endif

// gRabbit includes
#include "grabbit_common.h"
#include "grabbit_animation_parser.h"
#include "grabbit_animation_iterator.h"
#include "grabbit_instructions.h"

// C++ Includes
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cctype>
#include <cstring>


//----------------------------------------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------------------------------------
GrabbitAnimationParser::GrabbitAnimationParser(const wxString &filename) :
  haseof(false), state(0), lines(1), errornumber(0), errorline(1), token(""), errmessage(""), c(' ')
{
  file.open( filename.c_str(),  std::ios::in );
  if ( !file ) {
    errmessage.Printf("Coldn't open file %s", filename.c_str());
    std::cerr << errmessage.c_str() << std::endl;
    state=1;
    SetError(1);
  } else {
    file.flags ( std::ios::skipws );
    state=0;
  }
}

//----------------------------------------------------------------------------------------------------------------------
// Destructor
//----------------------------------------------------------------------------------------------------------------------
GrabbitAnimationParser::~GrabbitAnimationParser()
{
}

//----------------------------------------------------------------------------------------------------------------------
// Try to open an animation file
//----------------------------------------------------------------------------------------------------------------------
bool GrabbitAnimationParser::Open(const wxString &filename)
{
  file.open( filename.c_str(),  std::ios::in );
  if ( !file ) {
    errmessage.Printf("Coldn't open file %s", filename.c_str());
    std::cerr << errmessage.c_str() << std::endl;
    state=1;
    SetError(1);
    return false;
  } 
  file.flags ( std::ios::skipws );
  state = errornumber = 0;
  lines = errorline = 1;
  haseof = false;
  return true;
}

//----------------------------------------------------------------------------------------------------------------------
// Checks if the file was opened correctly
//----------------------------------------------------------------------------------------------------------------------
bool GrabbitAnimationParser::CheckOpened()
{
  if (state==0) return true;
  return false;
}

//----------------------------------------------------------------------------------------------------------------------
// Checks if an error has ocurred
//----------------------------------------------------------------------------------------------------------------------
int GrabbitAnimationParser::CheckError()
{
  if (!CheckOpened()) return 1;
  return errornumber;
}

//----------------------------------------------------------------------------------------------------------------------
// Checks if End of File has ocurred
//----------------------------------------------------------------------------------------------------------------------
bool GrabbitAnimationParser::CheckEOF()
{
  return haseof;
}

//----------------------------------------------------------------------------------------------------------------------
// Sets the error number
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::SetError(int errrr, unsigned long line)
{
  if (errornumber == 0) {
    errornumber = errrr;
    errorline = line;
    switch (errrr)
      {
      case 2: errmessage.Printf("Premature End of File"); break;
      case 3: errmessage.Printf("File Read Error"); break;
      case 4: errmessage.Printf("Double Quote [\"] expected"); break;
      case 5: errmessage.Printf("Opening Parenthesis [(] expected"); break;
      case 6: errmessage.Printf("Closing Parenthesis [)] expected"); break;
      case 7: errmessage.Printf("Semicolon [;] expected"); break;
      case 8: errmessage.Printf("Comma [,] expected"); break;
      case 9: errmessage.Printf("Number expected"); break;
      default: errmessage.Printf("Unknown error"); break;
      }
  } else if (errrr == 0) {
    errornumber = 0;
    errorline = 0;
    errmessage.clear();
  } else if (errrr == 2) {
    haseof = true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
// Jumps white spaces
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::JumpWhiteSpaces()
{
  do {
    if ( std::isspace( c ) ) { // Jump white spaces
      if (c == '\n') ++lines;
      continue;
    }
    break;
  } while ( file.get ( c ) );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a string into 'token' based on the folowwing Regular Expression: [A-Za-z]*
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadStr()
{
  token.clear();
  this->JumpWhiteSpaces();
  do {
    if ( !std::isalnum( c ) ) return;
    token << c;
  } while ( file.get ( c ) );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a string inside double quotes
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadStrQuote()
{
  token.clear();
  this->JumpWhiteSpaces();
  if (c != '"') return;
  while ( file.get ( c ) ) {
    if (c != '"') {
      if (c == '\n') break;
      token << c;
      continue;
    }
    break;
  }
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
  if (c != '"') SetError(4, lines);
  file.get( c );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a number
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadNumber()
{
  token.clear();
  this->JumpWhiteSpaces();
  do {
    if ( std::isdigit(c) ) {
      token << c;
      continue;
    }
    break;
  } while ( file.get ( c ) );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
  if (token.size() == 0)
    SetError(9,lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read an Open Parenthesis
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadOpenParenthesis()
{
  this->JumpWhiteSpaces();
  if (c != '(') SetError(5, lines);
  file.get ( c );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a Close Parenthesis
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadCloseParenthesis()
{
  this->JumpWhiteSpaces();
  if (c != ')') SetError(6, lines);
  file.get ( c );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a Semicolon
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadSemicolon()
{
  this->JumpWhiteSpaces();
  if (c != ';') SetError(7, lines);
  file.get ( c );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a Comma
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadComma()
{
  this->JumpWhiteSpaces();
  if (c != ',') SetError(8, lines);
  file.get ( c );
  if (file.eof())
    SetError(2, lines);
  else if (!file.good())
    SetError(3, lines);
}

//----------------------------------------------------------------------------------------------------------------------
// Read an "(r,g,b)" sequence (without double quotes)
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadRGB(unsigned char &r, unsigned char &g, unsigned char &b)
{
  unsigned long tmp;
  this->ReadOpenParenthesis();
  this->ReadNumber();
  token.ToULong( &tmp );
  r = (unsigned char)tmp;
  this->ReadComma();
  this->ReadNumber();
  token.ToULong( &tmp );
  g = (unsigned char)tmp;
  this->ReadComma();
  this->ReadNumber();
  token.ToULong( &tmp );
  b = (unsigned char)tmp;
  this->ReadCloseParenthesis();
}

//----------------------------------------------------------------------------------------------------------------------
// Prints an Error message
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::GetError(wxString &err)
{
  err.Printf("Error #%d: %s on line %ld", errornumber, errmessage.c_str(), errorline);
}

//----------------------------------------------------------------------------------------------------------------------
// Read a new animation instruction
//----------------------------------------------------------------------------------------------------------------------
void GrabbitAnimationParser::ReadInstruction(GrabbitAnimationIterator *it)
{
  GrabbitInstructions *instr;

  this->ReadStr();
  this->ReadOpenParenthesis();
  if (this->CheckError() != 0) return;
  
  if ( std::strcmp(token.c_str() , "LoadGraph") == 0 ) { //LoadGraph(string);
    this->ReadStrQuote();
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl; // DEBUG
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::LoadGraph );
      instr->SetName( token );
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "SaveGraph") == 0 ) { //SaveGraph(string);
    this->ReadStrQuote();
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::SaveGraph );
      instr->SetName( token );
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "PaintVertex") == 0 ) { //PaintVertex(name,(r,g,b));
    unsigned char r, g, b;
    wxString tmp("");
    
    this->ReadStr();
    tmp = token;
    this->ReadComma();
    this->ReadRGB(r, g, b);
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::PaintVertex );
      instr->SetName( tmp );
      instr->SetColor( r, g, b );
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "PaintArc") == 0 ) { //PaintArc(name,(r,g,b));
    unsigned char r, g, b;
    unsigned long pen_size=2;
    wxString tmp("");
    
    this->ReadStr();
    tmp = token;
    this->ReadComma();
    // -----------------------------------
    // Patch to change line size
    this->ReadNumber();
    token.ToULong( &pen_size );
    this->ReadComma();
    // -----------------------------------
    this->ReadRGB(r, g, b);
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError() != 0 && CheckError() != 2) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::PaintArc );
      instr->SetName( tmp );
      instr->SetColor( r, g, b );
      // -----------------------------------
      // Patch to change line size
      instr->SetPenSize( pen_size );
      // -----------------------------------
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "ChangeVertexText") == 0 ) { //ChangeVertexText(name, field, string, (r,g,b));
    unsigned char r, g, b;
    wxString name(""), field(""), str("");
    
    this->ReadStr();
    name = token;
    this->ReadComma();
    this->ReadStr();
    field = token;
    this->ReadComma();
    this->ReadStrQuote();
    str = token;
    this->ReadComma();
    this->ReadRGB(r, g, b);
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::ChangeVertexText );
      instr->SetName( name );
      instr->SetLabel( str );
      instr->SetColor( r, g, b );
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "ChangeArcText") == 0 ) { //ChangeArcText(name, field, string, (r,g,b));
    unsigned char r, g, b;
    wxString name(""), field(""), str("");
    
    this->ReadStr();
    name = token;
    this->ReadComma();
    this->ReadStr();
    field = token;
    this->ReadComma();
    this->ReadStrQuote();
    str = token;
    this->ReadComma();
    this->ReadRGB(r, g, b);
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::ChangeArcText );
      instr->SetName( name );
      instr->SetLabel( str );
      instr->SetColor( r, g, b );
      it->InsertIntruction( instr );
    }
    return;
  }
  if ( std::strcmp(token.c_str() , "Delay") == 0 ) { //Delay(time);
    unsigned long time;
    this->ReadNumber();
    token.ToULong( &time );
    this->ReadCloseParenthesis();
    if (CheckEOF()) SetError(10, lines);
    this->ReadSemicolon();
    // Creats here a new animation instruction
    if (CheckError()) {
      wxString TMP;
      GetError(TMP);
      std::cout << "Error: " << TMP.c_str() << std::endl;
    }
    if (!CheckError() || CheckError() == 2) {
      instr = new GrabbitInstructions();
      instr->SetType( GrabbitInstructions::Delay );
      instr->SetTime( time );
      it->InsertIntruction( instr );
    }
    return;
  }
}
